#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
## `velox.obj`

The `velox.obj` submodule provides all utilities, functionality, and logic around 
the instantiation, maintainence, and lifecycle of all `velox.obj.VeloxObject`
instances.
"""

from abc import ABCMeta, abstractmethod
import inspect
import logging
import os

from apscheduler.schedulers.background import BackgroundScheduler
from semantic_version import Version as SemVer, Spec as Specification
import six


from .exceptions import VeloxCreationError, VeloxConstraintError

from .filesystem import (find_matching_files, ensure_exists, stitch_filename,
                         get_aware_filepath)

from .tools import abstractclassmethod, timestamp, threaded, sha

logger = logging.getLogger(__name__)


def _default_prefix():
    vroot = os.environ.get('VELOX_ROOT')
    if vroot is None:
        vroot = os.path.abspath('.')
        logger.warning('falling back to {}, as no directory specified in '
                       'VELOX_ROOT'.format(vroot))
    return vroot


def _fail_bad_init(fn):
    """
    checks to make sure that an invoked objectmethod is invoked on a class 
    that has had it's superclass __init__() invoked.
    """
    from functools import wraps

    @wraps(fn)
    def _respect_reqs(cls, *args, **kw):
        if not hasattr(cls, '_parent_instantiated'):
            raise VeloxCreationError(
                'Object of type {} instantiated without '
                'call to constructor of super-class'.format(type(cls))
            )
        return fn(cls, *args, **kw)

    _respect_reqs.__doc__ = fn.__doc__
    return _respect_reqs


def _ensure_init(exclude):
    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)) and attr not in exclude:
                setattr(cls, attr, _fail_bad_init(getattr(cls, attr)))
        return cls
    return decorate


@_ensure_init(
    exclude=['__init__', 'load', '_load', '__del__', '__getstate__',
             '__setstate__', 'loadpath', 'clear_registered_names'],
)
class VeloxObject(object):
    """ `velox.obj.VeloxObject` provides a simple consistent way to handle saving,
    loading, versioning, and swapping binary objects.

    The core functionality depends on three main ideas.

    * A prefix, which tells Velox where to save blobs
    * A specifier, which tells Velox to search for particular substrings when
        handling aspiration
    * A version, which lets Velox know how much flexibility there is when
        loading new models

    The prefix is essentially the location where Velox will serialize objects.
    It can be on a filesystem or on s3. It can be set in two ways.

    * The `VELOX_ROOT` environment variable can be set to contain the
        default prefix.
    * During the invocation of `velox.obj.VeloxObject.save` or
        `velox.obj.VeloxObject.load`, a `prefix` keyword argument can be specified
        which allows an override of the default location.

    If the `prefix` keyword argument is not passed, we fall back to `VELOX_ROOT`
    if it is set, otherwise, we fall back to the current working directory.

    A prefix on a reachable filesystem needs nothing besides the path, and a
    prefix on S3 simply requires `s3://bucketName/foo/`.


    As an example, we could define the follwing model, inheriting of course from
    `velox.obj.VeloxObject`.

        #!python
        @register_model(
            registered_name='user_model', 
            version='1.1.4-alpha',
            version_constraints='>=1.0,<3.0'
        )
        class UserModel(VeloxObject):
            def __init__(self, user_model):
                super(UserModel, self).__init__()
                self._user_model = user_model
                self._etl = ETLPipeline()

            def _save(self, fileobject):
                pickle.dump(self, fileobject)

            @classmethod
            def _load(cls, fileobject):
                return pickle.load(fileobject)

            def predict(self, user):
                return self._user_model.predict(
                    self._etl.transform(user)
                )
    """

    __metaclass__ = ABCMeta

    # we dont want duplication of model names!
    _registered_object_names = []

    def __init__(self):
        """
        Base constructor for managed objects. 

        Raises:
        -------

        * `velox.exceptions.VeloxCreationError` if any type that inherits from
            this class fails to invoke this method (i.e., doesn't call 
            `super(NewModelClass, self).__init__()`)

        * `velox.exceptions.VeloxCreationError` if the class is defined without
            a wrapping call from `velox.obj.register_model`.
        """
        if not hasattr(self, '_registered_spec'):
            raise VeloxCreationError(
                'Managed Velox object instantiation failed due to missing '
                'registration decorator. The definition of all VeloxObject '
                'subclasses must be surrounded by the `register_model` '
                'decorator. Please consult the documentation for more details.'
            )
        self.__incr_underway = False
        self.__replacement = None
        self._scheduler = BackgroundScheduler()
        self._job_pointer = None
        self._current_sha = None
        self._parent_instantiated = True

    def __del__(self):
        if self._scheduler.state:
            self._scheduler.shutdown()
        if self._increment_underway:
            if not self.__replacement.done():
                self.__replacement.cancel()

    def __getstate__(self):
        # capture what is normally pickled
        state = self.__dict__.copy()
        del state['_scheduler']
        del state['_current_sha']
        return state

    def __setstate__(self, newstate):
        newstate['_scheduler'] = BackgroundScheduler()
        newstate['_current_sha'] = None
        self.__dict__.update(newstate)

    @property
    def current_sha(self):
        """
        Defines the current SHA1 of the filename that has most recently been
        loaded. If a file has never been loaded, this will be None.
        """
        return self._current_sha

    @current_sha.setter
    def current_sha(self, value):
        if self._current_sha is not None:
            raise ValueError('Cannot set a sha unless previous sha was None')
        self._current_sha = value

    @property
    def _needs_increment(self):
        if self.__replacement is None:
            return False
        return self.__replacement.done()

    @property
    def _increment_underway(self):
        return self.__incr_underway

    @classmethod
    def clear_registered_names(cls):
        """
        Unregisters all managed models.
        """
        logger.warning('Removing {} registered models. Proceed with caution.'
                       .format(len(cls._registered_object_names)))

        cls._registered_object_names = []

    @abstractmethod
    def _save(self, fileobject):
        raise NotImplementedError('super-class serialization not allowed')

    @abstractclassmethod
    def _load(cls, fileobject):
        raise NotImplementedError('super-class de-serialization not allowed')

    def save(self, prefix=None):
        """
        Saves the managed object instance using the user-defined method defined 
        in `_save`.

        Args:
        -----

        * `prefix (str)`: the prefix (can be on s3 or on a local filesystem) to 
            save the managed object to. If not passed will default to the
            value of the `VELOX_ROOT` env var if set, else, will fall back to
            the current working directory. 
        """

        outpath = self.savepath(prefix=prefix)
        logger.debug('assigned unique filepath: {}'.format(outpath))

        with get_aware_filepath(outpath, 'wb') as fileobject:
            self._save(fileobject)

        return outpath

    @classmethod
    def load(cls, prefix=None, specifier=None, skip_sha=None,
             local_cache_dir=None):
        """
        Loads a managed object instance using the user-defined method defined 
        in `_load`.

        Args:
        -----

        * `prefix (str)`: the prefix (can be on s3 or on a local filesystem) to
            load a managed object from. If not passed will default to the
            value of the `VELOX_ROOT` env var if set, else, will fall back to
            the current working directory.

        * `specifier (str)`: any substrings in the timestamp (as generated by
        `velox.tools.timestamp`) to explicitly search for.

        * `skip_sha (str)`: define a filename SHA1 to skip over.

        * `local_cache_dir (str)`: cache directory to dump a version of the 
            file from when loading. If the promotory version matches an 
            identifier in the cache, will load from the cache instead



        Raises:
        -------

        * `velox.exceptions.VeloxConstraintError` if we try to load from a SHA1 
            for which a skip was requested
        * `TypeError` if the user-defined `_load` function loads an object that 
            does not inherit from `velox.obj.VeloxObject`.

        """

        filepath = cls.loadpath(prefix=prefix, specifier=specifier)
        filesha = sha(get_filename(filepath))

        if skip_sha == filesha:
            raise VeloxConstraintError('found sha: {} when sha was explicitly '
                                       'blacklisted'.format(skip_sha))

        logger.debug('retrieving from filepath: {}'.format(filepath))

        if local_cache_dir is not None:
            ensure_exists(local_cache_dir)
            file_identifier = os.path.basename(filepath)

            local_copy = os.path.join(local_cache_dir, file_identifier)

            if os.path.isfile(local_copy):
                # if the file we want to load is on the local filesystem, load
                # from there instead
                filepath = local_copy
                logger.info('found target file in local cache. will load {} '
                            'from local copy'.format(file_identifier))
            else:
                logger.info('will dump to {} as cache copy'.format(local_copy))

        with get_aware_filepath(filepath, 'rb') as fileobject:
            obj = cls._load(fileobject)
            if not issubclass(type(obj), VeloxObject):
                raise TypeError('loaded object of type {} must inherit from '
                                'VeloxObject'.format(cls))
            obj.current_sha = filesha

            if local_cache_dir is not None and not os.path.isfile(local_copy):
                logger.info('dumping pulled copy to local filesystem')
                fileobject.seek(0)

                with open(local_copy, 'wb') as fp:
                    fp.write(fileobject.read())

                logger.info('cache op successful')

        return obj

    def _increment(self):
        replacement = self.__replacement.result()

        if self._current_sha != replacement.current_sha:
            logger.debug('will aspire to new version')
            # logger.debug('current sha: {}'.format(self._current_sha))
            logger.debug('current sha: {}'.format(self.current_sha))
            logger.debug('    new sha: {}'.format(replacement.current_sha))

            for k, v in replacement.__dict__.items():
                if (k not in {'_scheduler', '_job_pointer'}) and \
                        (not k.startswith('__')):
                    self.__dict__[k] = v
        else:
            logger.debug('found matching sha: {}'.format(self._current_sha))
            logger.debug('will skip increment'.format(self._current_sha))

        self.__incr_underway = False
        self.__replacement = None

    @threaded
    def __load_async(self, prefix, specifier, skip_sha):
        logger.debug('specifying a skip_sha = {}'.format(skip_sha))

        newobj = self.__class__.load(prefix, specifier, skip_sha=skip_sha)
        self.__incr_underway = False
        return newobj

    def __reload(self, prefix, specifier):
        self.__incr_underway = True

        try:
            self.__replacement = self.__load_async(prefix, specifier, None)
            # self.current_sha)
            self._increment()

        except VeloxConstraintError as ve:
            logger.debug('reload skipped. message: {}'.format(ve.args[0]))

    def reload(self, prefix=None, specifier=None, scheduled=False,
               **interval_trigger_args):
        """
        Defines the scheme by which to reload (hot swap) in-place. A scheduled 
        reload can be canceled with a call to 
        `velox.obj.VeloxObject.cancel_scheduled_reload`.

        Args:
        -----

        * `prefix (str)`: the prefix (can be on s3 or on a local filesystem) to
            load a managed object from. If not passed will default to the
            value of the `VELOX_ROOT` env var if set, else, will fall back to
            the current working directory.

        * `specifier (str)`: any substrings in the timestamp (as generated by
        `velox.tools.timestamp`) to explicitly search for.

        * `scheduled (bool)`: whether or not to run this as a scheduled and
            seperate threaded process (`True`) or to simply to an in-place swap
            (`False`). Only `scheduled=True` can guarantee zero-downtime.

        * `interval_trigger_args`: additional arguments to pass the the
            `BackgroundScheduler` object. Most commonly, you can pass something
            like `minutes=2` to schedule a poll to the prefix location every
            two minutes.

        Raises:
        -------

        * `ValueError` if you attempt to schedule a reload task when one is 
            already specified
        """

        if scheduled:
            if not self._scheduler.state:
                self._scheduler.start()
            if self._job_pointer is not None:
                raise ValueError('Found already-running job: '
                                 '{}'.format(self._job_pointer.id))
            logger.debug('scheduling with config: '
                         '{}'.format(interval_trigger_args))

            self._job_pointer = self._scheduler.add_job(
                func=self.__reload,
                args=(prefix, specifier),
                trigger='interval',
                max_instances=1,
                **interval_trigger_args
            )
            logger.debug('launched job {}: {}'
                         .format(self._job_pointer.id, self._job_pointer))

        else:
            logger.debug('initializing unscheduled async reload')
            self.__reload(prefix, specifier)

    def cancel_scheduled_reload(self):
        """
        Cancels a scheduled reload background task started through a call to 
        `velox.obj.VeloxObject.reload`.

        Raises:
        -------

        * `ValueError` if no jobs are available to cancel. 
        """
        if self._job_pointer is None:
            raise ValueError('no available job to cancel.')

        self._scheduler.remove_all_jobs()
        self._job_pointer = None

    def _register_name(self, name):
        self.__registered_name = name

    @property
    def registered_name(self):
        """
        Defines the registered name (i.e., from the `velox.obj.register_model` 
        decorator) of the class.
        """
        try:
            return self.__registered_name
        except NameError:
            raise VeloxCreationError('Usage of unregistered model')

    def formatted_filename(self):
        """
        Returns the formatted filename (without prefix) of the class, in the 
        format `'{timestamp}_{name}.vx'`.
        """
        return '{timestamp}_{name}.vx'.format(
            timestamp=timestamp(),
            name=self.registered_name
        )

    def savepath(self, prefix=None):
        """
        Stitches a prefix and a filename from 
        `velox.obj.VeloxObject.formatted_filename` to form a path to save a model 
        to.

        Args:
        -----

        * `prefix (str)`: the prefix (can be on s3 or on a local filesystem) to 
            save the managed object to. If not passed will default to the
            value of the `VELOX_ROOT` env var if set, else, will fall back to
            the current working directory. 

        Returns:
        --------

        A full filename where a model can be saved to.
        """
        if prefix is None:
            prefix = _default_prefix()
            logger.debug('No prefix specified. Falling back to '
                         'default at: {}'.format(prefix))

        ensure_exists(prefix)

        logger.debug('Prefix specified at: {}'.format(prefix))

        return stitch_filename(prefix, self.formatted_filename())

    @classmethod
    def loadpath(cls, prefix=None, specifier=None):
        """
        Determined the file to load from given the `prefix`, the `specifier`, 
        and any version constraint information from `velox.obj.register_model`. 
        Will always return the most recently created file that satisfies all 
        constraints.

        Args:
        -----

        * `prefix (str)`: the prefix (can be on s3 or on a local filesystem) to 
            load a managed object from. If not passed will default to the
            value of the `VELOX_ROOT` env var if set, else, will fall back to
            the current working directory.

        * `specifier (str)`: any substrings in the timestamp (as generated by 
        `velox.tools.timestamp`) to explicitly search for. 

        Returns:
        --------

        A full filename where a model can be loaded from.


        Raises:
        -------

        * `velox.exceptions.VeloxConstraintError` if no matching candidates 
            are found from the specified prefix subject to specified constraints.
        """

        if prefix is None:
            prefix = _default_prefix()
            logger.debug('No prefix specified. Falling back to '
                         'default at: {}'.format(prefix))

        v = SemVer(cls.__registered_name.split('_')[-1][1:])
        sortkey = cls.__registered_name.replace(str(v), '*')

        if specifier is None:
            specifier = '*_{}.vx'.format(sortkey)
        else:
            specifier = '*{}*_{}.vx'.format(specifier, sortkey)

        logger.info('Searching for matching file in {} with specifier {}'
                    .format(prefix, specifier))

        filelist = find_matching_files(prefix, specifier)

        if not filelist:
            raise VeloxConstraintError(
                'No files matching pattern {specifier} '
                'found in {prefix}'.format(specifier=specifier, prefix=prefix)
            )

        if cls._version_spec is not None:
            logger.debug('matching version requirements: '
                         '{}'.format(cls._version_spec))

            version_identifiers = list(map(get_semver, filelist))
            best_match = cls._version_spec.select(version_identifiers)
            logger.debug('found version to aspire to: {}'.format(best_match))

            if best_match is None:
                raise VeloxConstraintError(
                    'No files matching version requirements {} were '
                    'found'.format(cls._version_spec)
                )

            filelist = [
                fp for fp, v in zip(filelist, version_identifiers)
                if v == best_match
            ]

        if len(filelist) > 1:
            logger.warn('Found {} files matching. Selecting most '
                        'recent by filename timestamp'.format(len(filelist)))

        logger.info('will load from {}'.format(filelist[0]))

        return stitch_filename(prefix, filelist[0])


def _zero_downtime(fn):
    from functools import wraps

    @wraps(fn)
    def _respect_reload(cls, *args, **kw):
        if cls._needs_increment:
            logger.info('model version increment needed')
            cls._increment()
        # else:
        return fn(cls, *args, **kw)

    _respect_reload.__doc__ = fn.__doc__
    return _respect_reload


class register_model(object):

    """
    Wraps the `velox.obj.VeloxObject` subclass definitions to ensure all 
    necessary ancillary metadata is added. Allows for versioning and constraint
    definition for reloading procedure.

    For example, we could define the follwing model, inheriting of course from
    `velox.obj.VeloxObject`.

        #!python
        @register_model(
            registered_name='price_model', 
            version='0.2.1-rc1',
            version_constraints='>=0.1.0,<0.3.0,!=0.2.0'
        )
        class PriceModel(VeloxObject):
            def __init__(self, price_model):
                super(VeloxObject, self).__init__()
                self._price_model = price_model
                self._etl = ETLPipeline()

            def _save(self, fileobject):
                pickle.dump(self, fileobject)

            @classmethod
            def _load(cls, fileobject):
                return pickle.load(fileobject)

            def predict(self, X):
                return self._price_model.predict(
                    self._etl.transform(X)
                )
    """

    def __init__(self, registered_name, version='0.1.0-alpha',
                 version_constraints=None):
        """ Decorates an object with the required attributes to be managed by 
        Velox. Adds zero-downtime reloading to all non-velox-managed
        functionality.

        Args:
        -----

        * `registered_name (str)`: registration name for the class.
        * `version (str)`: a Sem Ver string specifying this classes version. 
        * `version_constraints (str | list)`: a Sem Ver version constraint 
            string  or list of strings specifying versioning restrictions for 
            loading.

        Raises:
        -------

        * `ValueError` if an invalid SemVer string is passed to either the 
            `version` or `version_constraints` keyword arguments. 

        * (on `__call__` invocation) `velox.exceptions.VeloxCreationError` if 
            `'{registered_name}_v{version}'` is not globally unique.

        * (on `__call__` invocation) `TypeError` if the defined class does not 
            inherit from `velox.obj.VeloxObject`.
        """

        try:
            registered_name = '{}_v{}'.format(registered_name, SemVer(version))
        except ValueError:
            raise ValueError('Invalid SemVer string: {}'.format(version))

        if registered_name in VeloxObject._registered_object_names:
            raise VeloxCreationError('Already a registered class named {}'
                                     .format(registered_name))

        VeloxObject._registered_object_names.append(registered_name)
        self._registered_name = registered_name

        if version_constraints is not None:
            if isinstance(version_constraints, six.string_types):
                self.version_specification = Specification(version_constraints)
            else:
                self.version_specification = Specification(*version_constraints)
        else:
            self.version_specification = None

    def __call__(self, cls):
        if not issubclass(cls, VeloxObject):
            raise TypeError('Type {} must inherit from '
                            'VeloxObject'.format(cls))
        if hasattr(cls, '_VeloxObject__registered_name'):
            raise VeloxCreationError('Class already registered!')
        setattr(cls, '_VeloxObject__registered_name',
                self._registered_name)

        setattr(cls, '_version_spec',
                self.version_specification)

        setattr(cls, '_registered_spec', True)

        reserved_attr = {
            'save',
            'reload',
            'cancel_scheduled_reload',
            'formatted_filename',
            'savepath'
        }

        for attr in cls.__dict__:
            fn = getattr(cls, attr)

            logger.debug('considering {}'.format(attr))

            if callable(fn) and inspect.getargspec(fn).args[0] == 'self':
                if not attr.startswith('_') and attr not in reserved_attr:

                    logger.info('adding zero-reload downtime for method {}'
                                .format(attr))

                    setattr(cls, attr, _fail_bad_init(_zero_downtime(fn)))

        return cls


def get_prefix(filepath):
    """
    From a `filepath`, will return the `prefix`
    """
    return os.path.split(filepath)[0]


def get_filename(filepath):
    """
    From a `filepath`, will return the filename **without** the `.vx` extension
    """
    return os.path.splitext(os.path.split(filepath)[-1])[0]


def _get_naming_info(filepath):
    return get_filename(filepath).split('_')


def get_registration_name(filepath):
    """
    Get the name passed to the `registered_name` keyword argument of the 
    `velox.obj.register_model` decorator.
    """
    return _get_naming_info(filepath)[1]


def get_specifier(filepath):
    """
    Gets the unique time identifier associated with filename.
    """
    return _get_naming_info(filepath)[0]


def get_semver(filepath):
    """
    Returns a `Version` object with the version of the passed-in filename 
    from the `semantic_version` package.
    """
    return SemVer(_get_naming_info(filepath)[2][1:])


def available_models():
    """
    Get a list of all available models (i.e., those registered with 
    `velox.obj.register_model`)
    """
    return VeloxObject._registered_object_names



import pytest

from backports.tempfile import TemporaryDirectory

import numpy as np
from velox.wrapper import SimplePickle, SimpleKeras


def build_test_data():

    X = np.random.normal(0, 1, (1000, 500))
    X_test = np.random.normal(0, 1, (1000, 500))
    y = X.dot([1, 2, 3, 1, 5] * 100)
    y += np.random.normal(0, 5, y.shape)

    return X, X_test, y


def test_simplepickle():
    from sklearn.linear_model import SGDRegressor

    X, X_test, y = build_test_data()

    clf = SGDRegressor(n_iter=20).fit(X, y)

    y_orig = clf.predict(X_test)

    with TemporaryDirectory() as d:
        SimplePickle(clf).save(prefix=d)

        model = SimplePickle.load(prefix=d)

    assert np.allclose(y_orig, model.predict(X_test))


def test_simplekeras():

    from keras.layers import Dense
    from keras.models import Sequential

    X, X_test, y = build_test_data()

    net = Sequential([
        Dense(20, activation='relu', input_dim=500),
        Dense(1)
    ])

    net.compile('adam', 'mse')

    net.fit(X, y, epochs=1)

    y_orig = net.predict(X_test)

    with TemporaryDirectory() as d:
        SimpleKeras(net).save(prefix=d)

        model = SimpleKeras.load(prefix=d)

    assert np.allclose(y_orig, model.predict(X_test))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
## `velox.filesystem`

The `velox.filesystem` submodule provides support utilities related to 
managing files (and by extension, S3) to the Velox ecosystem.
"""

from contextlib import contextmanager
import errno
import fnmatch
from glob import glob
import logging
import os
from tempfile import mkstemp

logger = logging.getLogger(__name__)


def find_matching_files(prefix, specifier):
    """
    Searches for files matching the `specifier` at the `prefix` location 
    (which can be on S3).
    """
    if not is_s3_path(prefix):
        logger.debug('Searching on filesystem')
        filelist = sorted(glob(os.path.join(os.path.abspath(prefix), specifier)))
    else:
        import boto3
        S3 = boto3.Session().resource('s3')

        bucket, key = parse_s3(prefix)

        logger.debug('searching in bucket s3://{} with '
                     'pfx key = {}'.format(bucket, key))

        objs_iterator = S3.Bucket(bucket).objects.filter(Prefix=key)

        filelist = sorted([
            os.path.basename(obj.key) for obj in objs_iterator
            if fnmatch.fnmatch(os.path.split(obj.key)[-1], specifier)
        ])

    return filelist[::-1]


def stitch_filename(prefix, filename):
    if is_s3_path(prefix):
        if prefix.endswith('/'):
            prefix = prefix[:-1]

        if not filename.startswith('/'):
            filename = '/' + filename

        return prefix + filename

    return os.path.join(prefix, filename)


def ensure_exists(prefix):
    """
    Safely ensures that the specified `prefix` exists. If `prefix` would point
    to a location on a reachable file system, it will safely create the 
    necessary directory path respecting race conditions. If `prefix` would 
    point to S3, it creates the bucket, if one doesn't already exist.
    """

    if not is_s3_path(prefix):
        logger.debug('Safely ensuring {} exists.'.format(prefix))
        safe_mkdir(prefix)
    else:
        import boto3

        logger.info('Prefix {} will be on S3'.format(prefix))

        bucket, key = parse_s3(prefix)

        logger.debug('S3 bucket = {}'.format(bucket))
        logger.debug('S3 key = {}'.format(key))

        SESSION = boto3.Session()

        S3 = SESSION.resource('s3')

        if bucket in {_.name for _ in S3.buckets.iterator()}:
            logger.debug('bucket already exists')
        else:
            logger.warn('bucket does not exist. Creating it...')
            S3.create_bucket(
                Bucket=bucket,
                CreateBucketConfiguration={
                    'LocationConstraint': SESSION.region_name
                }
            )


def safe_mkdir(path):
    '''
    Safe mkdir (i.e., don't create if already exists, 
    and no violation of race conditions)
    '''
    try:
        logger.debug('safely making (or skipping) directory: {}'.format(path))
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise exception


def is_s3_path(pth):
    return pth.startswith('s3://')


def parse_s3(pth):
    if not pth.startswith('s3://'):
        raise ValueError('{} is not a valid s3 path.'.format(pth))
    pth = pth.replace('s3://', '')
    split = pth.split(os.sep)
    # this gets back bucket, key
    return split[0], os.sep.join(split[1:])


@contextmanager
def get_aware_filepath(path, mode='w', session=None):
    """ context handler for dealing with local fs and remote (S3 only...)

    Args:
    -----
    * `path (str)`: path to object you wish to write. Can be 
            either `/path/to/desired/file.fmt`, or 
            `s3://myBucketName/this/is/a.key`

    * `mode (str)`: one of {rb, wb, r, w}

    * `session (None | boto3.Session)`: can pass in a custom boto3 session 
        if need be

    Example:
    --------

        #!python
        with get_aware_filepath('s3://bucket/file.txt', 'wb') as f:
            f.write('foobar')

        with get_aware_filepath('s3://bucket/file.txt', 'rb') as f:
            assert f.read() == 'foobar'

    Raises:
    -------

    * `ValueError` if `mode` is not one of the allowed modes.


    """

    if mode not in {'rb', 'wb', 'r', 'w'}:
        raise ValueError('mode must be one of {rb, wb, r, w}')

    if not is_s3_path(path):
        logger.debug('opening file = {} on local fs'.format(path))

        with open(path, mode) as f:
            yield f

        logger.debug('successfully closed session with file = {}'.format(path))

    else:

        if session is None:
            import boto3
            session = boto3.Session()

        S3 = session.resource('s3')

        fd, temp_fp = mkstemp(suffix='.tmpfile', prefix='s3_tmp', text=False)

        bucket, key = parse_s3(path)

        logger.debug('detected bucket = {}, key = {}, mode = {}'.format(
            bucket, key, mode))

        if mode in {'rb', 'r'}:
            logger.debug('initiating download to tempfile')
            S3.Bucket(bucket).download_file(key, temp_fp)
            logger.debug('download to tempfile successful')

        with open(temp_fp, mode) as f:
            logger.debug('yielding {} with mode {}'.format(temp_fp, mode))
            yield f
            logger.debug('closing {}'.format(temp_fp))

        if mode in {'wb', 'w'}:
            logger.debug('uploading {} to bucket = {} with key = {}'.format(
                temp_fp, bucket, key))
            S3.Bucket(bucket).upload_file(temp_fp, key)

        logger.debug('removing temporary allocations')
        os.close(fd)
        os.remove(temp_fp)

        logger.debug('cleaned up, releasing')

__all__ = ['get_aware_filepath', 'ensure_exists']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
## `velox.tools`

The `velox.tools` submodule provides general support utilities to the Velox 
package.
"""

from __future__ import unicode_literals

import datetime
from hashlib import md5
import six
from threading import Thread

from concurrent.futures import Future

from builtins import bytes


def sha(s):
    """
    get a simple, potentially value-inconsistent SHA of a python object.
    """
    m = md5()
    if isinstance(s, six.string_types):
        m.update(bytes(s, 'utf-8'))
    else:
        m.update(s.__repr__())
    return m.hexdigest()


def timestamp():
    """ 
    Returns a string of the form YYYYMMDDHHmmSSCCCCCC, where 

    * `YYYY` is the current year
    * `MM` is the current month, with zero padding to the left
    * `DD` is the current day, zero padded on the left
    * `HH`is the current hour, zero padded on the left
    * `mm`is the current minute, zero padded on the left
    * `CCCCCC`is the current microsecond as a decimal number, 
        zero padded on the left
    """
    return datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S%f')


class abstractclassmethod(classmethod):

    # this hack allows us to enforce the ABC implementation of a classmethod
    __isabstractmethod__ = True

    def __init__(self, callable):
        callable.__isabstractmethod__ = True
        super(abstractclassmethod, self).__init__(callable)


def call_with_future(fn, future, args, kwargs):
    try:
        result = fn(*args, **kwargs)
        future.set_result(result)
    except Exception as exc:
        future.set_exception(exc)


def threaded(fn):
    """
    A simple decorator that allows a function to be called with its return 
    value given as a `Future` object.

    Example:
    --------

        #!python
        @threaded
        def fn(x):
            return x ** 2

        a = fn(6)
        assert a.result() == 36
    """

    def wrapper(*args, **kwargs):
        future = Future()
        Thread(target=call_with_future,
               args=(fn, future, args, kwargs)).start()
        return future
    return wrapper

__all__ = ['threaded', 'timestamp']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

u"""

<img src="https://github.com/lukedeo/Velox/raw/master/img/velox-logo.png" width=37% align="right" />

# Welcome to Velox!

[![Build Status](https://travis-ci.org/lukedeo/Velox.svg?branch=master)](https://travis-ci.org/lukedeo/Velox)
[![Coverage Status](https://coveralls.io/repos/github/lukedeo/Velox/badge.svg?branch=master)](https://coveralls.io/github/lukedeo/Velox?branch=master)
[![PyPI version](https://badge.fury.io/py/Velox.svg)](https://badge.fury.io/py/Velox)

Deploying and managing live machine learning models is difficult. It involves a mix of handling model versioning, hot-swapping new versions and determining version constraint satisfaction on-the-fly, and managing binary file movement either on a networked or local file system or with a cloud storage system like S3. Velox can handle this for you with a simple base class enforcing opinionated methods of handling the above problems.

Velox provides two main utilities:

* Velox abstracts the messiness of consistent naming schemes and handling saving and loading requirements for a filesystem and for other forms of storage.
* Velox allows the ability to do a model / blob hotswap in-place for a new binary version.

## Requirements
---

~~Velox currently only supports Python 2.7, but **we would love contributions towards Python 3 support** 😁~~

**Velox supports Python 2 and Python 3**

The main requirements are `apscheduler` for scheduling hot-swaps, `semantic_version` for version sanity, and the `futures` Python 2.7 backport. If you want to be able to work with S3, you'll need `boto3` (and a valid and properly set up AWS account).

To run the tests, you'll need the brilliant `moto` library, the `backports.tempfile` library for Python 2.7 compatibility, and `Keras` and `sckit-learn`.

For logging, simply grab the Velox logger by the `velox` handle.

You can install Velox using `pip install velox`, 

## `VeloxObject` Abstract Base Class
---

Functionality is exposed using the `VeloxObject` abstract base class (ABC). A subclass of a `velox.obj.VeloxObject` needs to implement three things in order for the library to know how to manage it.

* Your class must be defined with a `velox.obj.register_model` decorator around it.
* Your class must implement a `_save` object method that takes as input a file object and does whatever is needed to save the object.
* Your class must implement a `_load` class method (with the `@classmethod` decorator) that takes as input a file object and reconstructs and returns an instance of your class.

This allows you to abstract away much of the messiness in bookkeeping.

Here is a simple example showing all required components.

<!--begin_code-->
    #!python
    @register_model(
        registered_name='foobar',
        version='0.1.0-alpha',
        version_constraints='>=0.1.0,<0.2.0'
    )
    class ChurnModel(VeloxObject):
        def __init__(self, submodel):
            super(VeloxObject, self).__init__()
            self._submodel = submodel

        def _save(self, fileobject):
            pickle.dump(self, fileobject)

        @classmethod
        def _load(cls, fileobject):
            return pickle.load(fileobject)

        def predict(self, X):
            return self._submodel.predict(X)
<!--end_code-->


Here is a full example using [`gensim`](https://github.com/RaRe-Technologies/gensim) to build a topic model and keep track of all the necessary ETL-type objects that follow:

<!--begin_code-->
    #!python


    import os
    import shutil
    import tarfile
    import tempfile

    from gensim.corpora import Dictionary
    from gensim.models.ldamulticore import LdaMulticore
    from spacy.en import English

    from velox import VeloxObject, register_model


    nlp = English()


    @register_model('lda', '0.2.1')
    class LDAModel(VeloxObject):

        def __init__(self, n_components=50):
            super(LDAModel, self).__init__()

            self._n_components = n_components

        @staticmethod
        def tokenize(text):
            return [t.text.lower()
                    for t in nlp(unicode(text))
                    if t.text.strip()]

        def fit(self, texts, passes=5, n_workers=1,
                no_below=5, no_above=0.2):
            tokenized = map(self.tokenize, texts)

            self._dictionary = Dictionary(tokenized)
            self._dictionary.filter_extremes(
                no_below=no_below,
                no_above=no_above
            )

            self._lda = LdaMulticore(
                corpus=map(
                    self._dictionary.doc2bow,
                    tokenized
                ),
                workers=n_workers,
                num_topics=self._n_components,
                id2word=self._dictionary,
                passes=passes
            )

            return self

        def transform(self, texts):
            feats = (
                self._dictionary.doc2bow(self.tokenize(q))
                for q in texts
            )
            vecs = list(self._lda[feats])

            X = np.zeros((len(vecs), self._lda.num_topics))
            for i, v in enumerate(vecs):
                ix, val = zip(*v)
                X[i][np.array(ix)] = val
            return X

        def _save(self, fileobject):
            with tarfile.open(fileobject.name, 'w:') as tf:
                with tempfile.NamedTemporaryFile() as tmp:
                    self._dictionary.save(tmp.name)
                    tf.add(tmp.name, 'dict.model')

                tmpdir = tempfile.mkdtemp()

                sp = os.path.join(tmpdir, 'lda.model')
                self._lda.save(sp)

                tf.add(tmpdir, 'lda.dir')

        @classmethod
        def _load(cls, fileobject):

            tmpdir = tempfile.mkdtemp()
            with tarfile.open(fileobject.name, 'r:*') as tf:
                tf.extractall(tmpdir)

            model = cls()

            _dictionary = Dictionary.load(
                os.path.join(tmpdir, 'dict.model')
            )
            _lda = LdaMulticore.load(
                os.path.join(tmpdir, 'lda.dir', 'lda.model')
            )

            setattr(model, '_dictionary', _dictionary)
            setattr(model, '_lda', _lda)

            setattr(model, '_n_components', model._lda.num_topics)
            shutil.rmtree(tmpdir)

            return model

<!--end_code-->

Voilà! Now, let's say you have a list of texts, and you wanted to train this 
model:

<!--begin_code-->
    #!python

    texts = [...]

    lda = LDAModel(128)
    lda.fit(texts)

    T = lda.transform(texts)

    lda.save('s3://my-ci-bucket/models/foo')
<!--end_code-->

Elsewhere, (i.e., a production server, etc.) you can load the latest model like
so:

<!--begin_code-->
    #!python
    production_lda = LDAModel.load('s3://my-ci-bucket/models/foo')
    T = production_lda.transform(...)
<!--end_code-->

In most environments, we would like the model to get hotswapped when a new model 
is uploaded to `s3`. Velox makes this easy! As long as the model stays in 
memory, we can use a async reload process to poll the `prefix` location for
updated models!

<!--begin_code-->
    #!python

    production_lda = LDAModel.load('s3://my-ci-bucket/models/foo')
    production_lda.reload(
        prefix='s3://my-ci-bucket/models/foo', 
        scheduled=True, 
        minutes=5
    )



    T = production_lda.transform(...)
<!--end_code-->


"""
import logging

# supress "No handlers could be found for logger XXXXX" error
logging.getLogger('velox').addHandler(logging.NullHandler())

__version__ = '0.2.0'

from .obj import VeloxObject, register_model

from . import filesystem
from . import exceptions
from . import tools
from . import obj
from . import wrapper

__all__ = ['filesystem', 'exceptions', 'tools', 'obj', 'wrapper']
